"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.CLIOptions = void 0;
const tslib_1 = require("tslib");
const FS = tslib_1.__importStar(require("fs"));
const OS = tslib_1.__importStar(require("os"));
const Path = tslib_1.__importStar(require("path"));
const clime_1 = require("clime");
const prompts_1 = tslib_1.__importDefault(require("prompts"));
const tiva_1 = require("tiva");
const main_1 = require("../main");
const shared_1 = require("../shared");
const JSON_CONFIG_INDENTATION = 2;
const WORKSPACE_PATH_DEFAULT = Path.resolve(OS.homedir(), '.config', 'makescript', 'agent');
const GENERATE_CONFIG_DEFAULT = false;
const CONFIG_FILE_NAME = 'agent.json';
class CLIOptions extends clime_1.Options {
}
tslib_1.__decorate([
    clime_1.option({
        flag: 'g',
        description: 'Whether only to generate makescript agent config file.',
        toggle: true,
        default: GENERATE_CONFIG_DEFAULT,
    }),
    tslib_1.__metadata("design:type", Boolean)
], CLIOptions.prototype, "generateConfigOnly", void 0);
tslib_1.__decorate([
    clime_1.option({
        flag: 'l',
        description: 'The makescript host join link.',
        // TODO: It will throw an error if this is not pass
        type: clime_1.Castable.CommaSeparatedStrings,
    }),
    tslib_1.__metadata("design:type", Object)
], CLIOptions.prototype, "joinLink", void 0);
tslib_1.__decorate([
    clime_1.option({
        flag: 'w',
        description: 'The path for makescript agent to work.',
        default: WORKSPACE_PATH_DEFAULT,
    }),
    tslib_1.__metadata("design:type", clime_1.Castable.Directory)
], CLIOptions.prototype, "workspace", void 0);
exports.CLIOptions = CLIOptions;
let default_1 = class default_1 extends clime_1.Command {
    async execute({ generateConfigOnly, workspace, joinLink: [joinLink] = [], }) {
        let configFilePath = Path.join(workspace.fullName, CONFIG_FILE_NAME);
        let configFileExists = FS.existsSync(configFilePath);
        if (configFileExists && generateConfigOnly) {
            console.warn(`The config file "${configFilePath}" has existed, so the new config has not generated.`);
            return;
        }
        if (!configFileExists) {
            let dirname = workspace.fullName;
            if (!FS.existsSync(dirname)) {
                FS.mkdirSync(dirname, { recursive: true });
            }
            let answer = await initialQuestions();
            let jsonConfig = {
                makescript: {
                    joinLink: answer.joinLink,
                    namespace: answer.namespace,
                },
                scripts: {
                    repoURL: answer.repoURL,
                    // if subPath is "", pass undefined instead.
                    path: answer.subPath || undefined,
                },
                proxy: undefined,
            };
            writeConfig(jsonConfig);
        }
        if (generateConfigOnly) {
            return;
        }
        let jsonConfig = readConfig();
        if (joinLink && jsonConfig.makescript.joinLink !== joinLink) {
            jsonConfig.makescript.joinLink = joinLink;
            writeConfig(jsonConfig);
        }
        let tiva = new tiva_1.Tiva({
            project: Path.join(__dirname, '../../../src/program'),
        });
        shared_1.logger.info('Checking config file ...');
        try {
            await tiva.validate({
                module: './config',
                type: 'JSONConfigFile',
            }, jsonConfig);
            await main_1.main(tiva, {
                ...jsonConfig,
                workspace: workspace.fullName,
                agentModule: undefined,
            });
        }
        catch (error) {
            if (error.diagnostics) {
                shared_1.logger.error(`Config file structure does not match:\n${error.diagnostics}`);
            }
            else {
                shared_1.logger.error(`Unknown error occurred:\n${error.message}`);
            }
            process.exit(1);
        }
        function writeConfig(config) {
            let jsonConfigText = JSON.stringify(config, undefined, JSON_CONFIG_INDENTATION);
            FS.writeFileSync(configFilePath, jsonConfigText);
        }
        function readConfig() {
            let jsonConfigText = FS.readFileSync(configFilePath).toString();
            return JSON.parse(jsonConfigText);
        }
        async function initialQuestions() {
            var _a;
            let promptObjects = [];
            if (!joinLink) {
                promptObjects.push({
                    type: 'text',
                    name: 'joinLink',
                    message: 'Please enter the makescript host join link',
                    validate: value => /^https?:\/\/.+$/.test(value),
                });
            }
            promptObjects.push(...[
                {
                    type: 'text',
                    name: 'namespace',
                    message: 'What namespace do you want to register as',
                },
                {
                    type: 'text',
                    name: 'repoURL',
                    message: 'Please enter the git repo url of the scripts',
                    validate: value => /^(https?:\/\/.+)|(\w+\.git)$/.test(value),
                },
                {
                    type: 'text',
                    name: 'subPath',
                    message: 'Please enter the path of the scripts definition in the repo',
                },
            ]);
            let answer = await prompts_1.default(promptObjects);
            // There is a bug (or unhandled behavior) with 'prompts'.
            // When user press CTRL + C , program will continue to execute with empty answers.
            // https://github.com/terkelg/prompts/issues/252
            if ((!joinLink && !answer.joinLink) ||
                !answer.namespace ||
                !answer.repoURL) {
                process.exit(0);
            }
            return {
                joinLink: joinLink !== null && joinLink !== void 0 ? joinLink : answer.joinLink,
                namespace: answer.namespace,
                repoURL: answer.repoURL,
                subPath: (_a = answer.subPath) !== null && _a !== void 0 ? _a : undefined,
            };
        }
    }
};
tslib_1.__decorate([
    clime_1.metadata,
    tslib_1.__metadata("design:type", Function),
    tslib_1.__metadata("design:paramtypes", [CLIOptions]),
    tslib_1.__metadata("design:returntype", Promise)
], default_1.prototype, "execute", null);
default_1 = tslib_1.__decorate([
    clime_1.command()
], default_1);
exports.default = default_1;
//# sourceMappingURL=default.js.map