"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.RunningService = void 0;
const tslib_1 = require("tslib");
const CP = tslib_1.__importStar(require("child_process"));
const FS = tslib_1.__importStar(require("fs"));
const OS = tslib_1.__importStar(require("os"));
const Path = tslib_1.__importStar(require("path"));
const bcrypt_1 = tslib_1.__importDefault(require("bcrypt"));
const rimraf_1 = tslib_1.__importDefault(require("rimraf"));
const uuid_1 = require("uuid");
const villa = tslib_1.__importStar(require("villa"));
const _utils_1 = require("../@utils");
const constants_1 = require("../constants");
const shared_1 = require("../shared");
const MAKESCRIPT_TMPDIR = Path.join(OS.tmpdir(), 'makescript-temp');
const OUTPUT_FLUSH_INTERVAL_TIME = 1000;
class RunningService {
    constructor(scriptService, socketService) {
        this.scriptService = scriptService;
        this.socketService = socketService;
        this.adapterMap = new Map();
    }
    async triggerHook(scriptName, hookName) {
        var _a;
        let definition = this.requireScriptDefinition(scriptName);
        let hookContent = (_a = definition.hooks) === null || _a === void 0 ? void 0 : _a[hookName];
        if (!hookContent) {
            throw new Error(`The hook "${hookName}" of script "${scriptName}" is not configured.`);
        }
        let cp = CP.exec(hookContent, {
            cwd: this.scriptService.scriptsBasePath,
            env: {
                ...process.env,
                ...this.scriptService.getEnvByScriptName(scriptName),
            },
        });
        await villa.awaitable(cp);
    }
    async runScript(argument) {
        let { name, parameters, resourcesBaseURL } = argument;
        shared_1.logger.info(`Running record "${argument.id}" of script "${argument.name}"`);
        let definition = this.requireScriptDefinition(name);
        await this.validatePassword(definition, argument);
        let adapter = this.requireAdapter(definition);
        let options = this.resolveOptions(definition);
        let resourcesPath = this.generateRandomResourcesPath();
        let [allowedParameters, deniedParameters] = validateParameters(parameters, definition);
        let { onOutput, done: onOutputDone } = this.getOnOutput(argument, false);
        let { onOutput: onError, done: onErrorDone } = this.getOnOutput(argument, true);
        let result = await adapter.runScript({
            repoPath: this.scriptService.scriptsBasePath,
            cwd: this.scriptService.scriptsPath,
            env: this.scriptService.getEnvByScriptName(name),
            definition,
            parameters: allowedParameters,
            options,
            resourcesPath,
            resourcesBaseURL,
            onOutput,
            onError,
        });
        let output = await onOutputDone();
        let error = await onErrorDone();
        await this.handleResources(argument, resourcesPath);
        shared_1.logger.info(`Complete running record "${argument.id}" of script "${argument.name}"`);
        return {
            name,
            displayName: definition.displayName,
            parameters,
            deniedParameters,
            result,
            output: {
                output,
                error,
            },
        };
    }
    registerAdapter(type, adapter) {
        this.adapterMap.set(type, adapter);
    }
    requireScriptDefinition(name) {
        let scriptDefinition = this.scriptService.getDefaultValueFilledScriptDefinitionByName(name);
        if (!scriptDefinition) {
            throw new Error(`Script definition "${name}" not found`);
        }
        return scriptDefinition;
    }
    async validatePassword(scriptDefinition, { password }) {
        if (!scriptDefinition.passwordHash) {
            return;
        }
        let result = await bcrypt_1.default.compare(password, scriptDefinition.passwordHash);
        if (!result) {
            throw new Error(`Password error: the password provided to running "${scriptDefinition.name}" is not match`);
        }
    }
    requireAdapter(scriptDefinition) {
        let adapter = this.adapterMap.get(scriptDefinition.type);
        if (!adapter) {
            // TODO:
            throw Error(`Adapter for script type "${scriptDefinition.type}" not found`);
        }
        return adapter;
    }
    resolveOptions(scriptDefinition) {
        if (!scriptDefinition.options) {
            return {};
        }
        return Object.fromEntries(Object.entries(scriptDefinition.options).map(([name, value]) => {
            if (value.type === 'value') {
                return [name, value.value];
            }
            else if (value.type === 'env') {
                let resolvedValue = process.env[value.env];
                if (!resolvedValue && value.required) {
                    // TODO:
                    throw new Error();
                }
                return [name, resolvedValue];
            }
            // TODO:
            throw new Error();
        }));
    }
    generateRandomResourcesPath() {
        return Path.join(OS.tmpdir(), 'makescript', 'agent', 'resources', uuid_1.v4());
    }
    async handleResources(argument, resourcesPath) {
        if (!FS.existsSync(resourcesPath)) {
            return;
        }
        shared_1.logger.info(`Transmitting resources for record "${argument.id}" of script "${argument.name}"`);
        let temporaryArchiveFilePath = Path.join(MAKESCRIPT_TMPDIR, `${uuid_1.v4()}.zip`);
        if (!FS.existsSync(MAKESCRIPT_TMPDIR)) {
            FS.mkdirSync(MAKESCRIPT_TMPDIR);
        }
        await _utils_1.zip(resourcesPath, temporaryArchiveFilePath);
        let buffer = await villa.async(FS.readFile)(temporaryArchiveFilePath);
        await this.socketService.makescriptRPC.updateResources(argument.id, buffer);
        await villa.async(rimraf_1.default)(temporaryArchiveFilePath);
    }
    getOnOutput(argument, error) {
        let output = '';
        let lastFlushedText;
        let flushOutput = () => {
            if (error) {
                return;
            }
            let textToFlush = getTextToFlush(output);
            if (!textToFlush || textToFlush === lastFlushedText) {
                return;
            }
            // TODO: Ensure sequence in time
            this.socketService.makescriptRPC
                .updateOutput(argument.id, textToFlush)
                .catch(shared_1.logger.error);
        };
        let timer = setInterval(() => {
            flushOutput();
        }, OUTPUT_FLUSH_INTERVAL_TIME);
        return {
            onOutput: data => {
                output += data;
            },
            done: async () => {
                clearInterval(timer);
                flushOutput();
                return output;
            },
        };
    }
}
exports.RunningService = RunningService;
function validateParameters(parameters, definition) {
    if (!definition.parameters) {
        return [{}, {}];
    }
    let { filteredParameters, missingParameters } = definition.parameters.reduce((reducer, parameterDefinition) => {
        let { filteredParameters, missingParameters } = reducer;
        let parameterName = typeof parameterDefinition === 'object'
            ? parameterDefinition.name
            : parameterDefinition;
        let parameterRequired = typeof parameterDefinition === 'object'
            ? parameterDefinition.required
            : false;
        let parameterValue = parameters[parameterName];
        let serializedValue = typeof parameterValue === 'object'
            ? JSON.stringify(parameterValue)
            : parameterValue;
        if (serializedValue !== undefined) {
            filteredParameters[parameterName] = serializedValue;
        }
        else {
            if (parameterRequired) {
                missingParameters.push(parameterName);
            }
        }
        return reducer;
    }, { filteredParameters: {}, missingParameters: [] });
    if (missingParameters.length) {
        throw new Error(`Missing command required parameters "${missingParameters.join('", "')}"`);
    }
    let deniedParameters = Object.fromEntries(Object.entries(parameters).filter(([key]) => !(key in filteredParameters)));
    return [filteredParameters, deniedParameters];
}
function getTextToFlush(text) {
    if (!text.includes(constants_1.OUTPUT_CLEAR_CHARACTER)) {
        return text;
    }
    let splitTexts = text.split(constants_1.OUTPUT_CLEAR_CHARACTER);
    return splitTexts[splitTexts.length - 1];
}
//# sourceMappingURL=running-service.js.map