"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
exports.ScriptService = void 0;
const tslib_1 = require("tslib");
const CP = tslib_1.__importStar(require("child_process"));
const FS = tslib_1.__importStar(require("fs"));
const Path = tslib_1.__importStar(require("path"));
const rimraf_1 = tslib_1.__importDefault(require("rimraf"));
const villa = tslib_1.__importStar(require("villa"));
const shared_1 = require("../shared");
const SCRIPTS_DIRECTORY_NAME = 'scripts';
const SCRIPTS_CONFIG_FILE_NAME_JSON = 'makescript.json';
const AGENT_MODULE_DEFAULT = './types';
class ScriptService {
    constructor(tiva, config) {
        this.tiva = tiva;
        this.config = config;
        this.ready = this.initialize();
    }
    get scriptsBasePath() {
        return Path.join(this.config.workspace, SCRIPTS_DIRECTORY_NAME);
    }
    get scriptsPath() {
        let scriptsBasePath = this.scriptsBasePath;
        let scriptsSubPath = this.config.scripts.path;
        return scriptsSubPath
            ? Path.join(scriptsBasePath, scriptsSubPath)
            : scriptsBasePath;
    }
    get scriptsDefinitionPath() {
        return Path.join(this.scriptsPath, SCRIPTS_CONFIG_FILE_NAME_JSON);
    }
    get briefScriptDefinitions() {
        let scriptsDefinition = this.scriptsDefinition;
        if (!scriptsDefinition) {
            return [];
        }
        return scriptsDefinition.scripts.map(scriptDefinition => convertScriptDefinitionToBriefScriptDefinition(fillScriptDefinitionDefaultValue(scriptDefinition, scriptsDefinition)));
    }
    async syncScripts() {
        var _a;
        shared_1.logger.info('Syncing scripts ...');
        try {
            if (FS.existsSync(this.scriptsBasePath)) {
                let cp = CP.spawn('git', ['remote', 'get-url', 'origin'], {
                    cwd: this.scriptsBasePath,
                });
                let remoteURL = '';
                cp.stdout.on('data', (buffer) => {
                    remoteURL += buffer.toString();
                });
                await villa.awaitable(cp);
                if (remoteURL.trim() !== this.config.scripts.repoURL.trim()) {
                    shared_1.logger.info('Scripts repo url changed, start to sync from the new url');
                    await villa.call(rimraf_1.default, this.scriptsBasePath);
                    await villa.awaitable(CP.spawn('git', [
                        'clone',
                        this.config.scripts.repoURL,
                        this.scriptsBasePath,
                    ]));
                }
                else {
                    await villa.awaitable(CP.spawn('git', ['pull'], {
                        cwd: this.scriptsBasePath,
                    }));
                }
            }
            else {
                await villa.awaitable(CP.spawn('git', [
                    'clone',
                    this.config.scripts.repoURL,
                    this.scriptsBasePath,
                ]));
            }
        }
        catch (error) {
            throw new Error(`Failed to sync scripts: ${error.message}`);
        }
        let scriptsDefinition = await this.parseScriptsDefinition();
        this.scriptsDefinition = scriptsDefinition;
        if ((_a = scriptsDefinition.hooks) === null || _a === void 0 ? void 0 : _a.install) {
            try {
                shared_1.logger.info('Initializing scripts ...');
                await villa.awaitable(CP.exec(scriptsDefinition.hooks.install, {
                    cwd: this.scriptsBasePath,
                }));
            }
            catch (error) {
                throw new Error(`Cannot to initial script repo with \`${scriptsDefinition.hooks.install}\`: ${error.message}`);
            }
        }
        shared_1.logger.info('Scripts initialized');
    }
    getDefaultValueFilledScriptDefinitionByName(name) {
        let scriptsDefinition = this.scriptsDefinition;
        if (!scriptsDefinition) {
            return;
        }
        let definition = scriptsDefinition.scripts.find(definition => definition.name === name);
        if (!definition) {
            return undefined;
        }
        return fillScriptDefinitionDefaultValue(definition, scriptsDefinition);
    }
    getEnvByScriptName(scriptName) {
        return {
            SCRIPT_NAME: scriptName,
            NAMESPACE: this.config.makescript.namespace,
        };
    }
    async initialize() {
        await this.syncScripts();
    }
    async parseScriptsDefinition() {
        var _a;
        if (!FS.existsSync(this.scriptsBasePath)) {
            throw new Error(`Scripts repo not cloned`);
        }
        if (!FS.existsSync(this.scriptsDefinitionPath)) {
            throw new Error('Scripts definition not found: \n' +
                `please ensure the definition file "${SCRIPTS_CONFIG_FILE_NAME_JSON}" is existing in scripts repo.`);
        }
        shared_1.logger.info('Checking scripts definition file ...');
        let scriptsDefinitionBuffer = await villa.async(FS.readFile)(this.scriptsDefinitionPath);
        let scriptsDefinitionContent = scriptsDefinitionBuffer.toString();
        let parsedDefinition = JSON.parse(scriptsDefinitionContent);
        try {
            await this.tiva.validate({
                module: (_a = this.config.agentModule) !== null && _a !== void 0 ? _a : AGENT_MODULE_DEFAULT,
                type: 'ScriptsDefinition',
            }, parsedDefinition);
        }
        catch (error) {
            if (error.diagnostics) {
                shared_1.logger.error(`The structure of the scripts definition file "${this.scriptsDefinitionPath}" not matched: \n` +
                    `    ${error.diagnostics}`);
                process.exit(1);
            }
            else {
                throw error;
            }
        }
        shared_1.logger.info('The scripts definition file is correct');
        return parsedDefinition;
    }
}
exports.ScriptService = ScriptService;
function fillScriptDefinitionDefaultValue(definition, scriptsDefinition) {
    var _a;
    return {
        ...definition,
        passwordHash: (_a = definition.passwordHash) !== null && _a !== void 0 ? _a : scriptsDefinition.passwordHash,
        manual: definition.manual === true,
        hooks: {
            ...scriptsDefinition.hooks,
            ...definition.hooks,
        },
    };
}
function convertScriptDefinitionToBriefScriptDefinition(definition) {
    var _a, _b;
    return {
        displayName: definition.displayName,
        name: definition.name,
        type: definition.type,
        manual: definition.manual === true,
        parameters: (_a = definition.parameters) !== null && _a !== void 0 ? _a : [],
        needsPassword: !!definition.passwordHash,
        hooks: {
            postscript: !!((_b = definition.hooks) === null || _b === void 0 ? void 0 : _b.postscript),
        },
    };
}
//# sourceMappingURL=script-service.js.map